# Copyright 2019 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Helper functions for visualization."""


def tfk_history(history, output_dir=None, **ax_set_kwargs):
  """Visualization of `tensorflow.keras.callbacks.History`,
  similar to `TensorBoard`, in train and validation.
  
  Args:
    history: `tensorflow.keras.callbacks.History`, the logs of
      training a `tensorflow.keras.Model`.
    output_dir: (optional) `str`, the directory name to
      store the figures.
  """
  import os
  import tensorflow as tf
  tfk = tf.keras
  import matplotlib.pyplot as plt
  if not isinstance(history, tfk.callbacks.History):
    raise TypeError("`history` was expected to be of type "
                    "`tensorflow.keras.callbacks.History`, "
                    "but {} was provided.".format(type(history)))
  for metric in [k for k in history.history.keys() if not "val_" in k]:
    fig, ax = plt.subplots()
    ax.plot(history.history.get(metric), label="train")
    ax.plot(history.history.get("val_{}".format(metric)), label="validation")
    ax.set(title=metric, xlabel="epochs", **ax_set_kwargs)
    ax.legend()
    fig.tight_layout()
    if isinstance(output_dir, str):
      os.makedirs(output_dir, exist_ok=True)
      fig.savefig(os.path.join(output_dir, "{}.pdf".format(metric)),
                  trasparent=True,
                  dpi=300,
                  format="pdf")
    fig.show()


def leaderboard(benchmark,
                results=None,
                output_dir=None,
                leaderboard_dir=None,
                **ax_set_kwargs):
  """Generates a leaderboard for all metrics in `benchmark`,
  by appending the (optional) `results`.
  
  Args:
    benchmark: `str`, the registerd name of `bdlb.Benchmark`.
    results: (optional) `dict`, dictionary of `pandas.DataFrames`
      with the results from a new method to be plotted against
      the leaderboard.
    leaderboard_dir: (optional) `str`, the path to the parent
      directory with all the leaderboard results.
    output_dir: (optional) `str`, the directory name to
      store the figures.
  """
  import os
  import pandas as pd
  import matplotlib.pyplot as plt
  from .constants import BDLB_ROOT_DIR
  COLORS = plt.rcParams['axes.prop_cycle'].by_key()['color']
  MARKERS = ["o", "D", "s", "8", "^", "*"]

  # The assumes path for stored baselines records
  leaderboard_dir = leaderboard_dir or os.path.join(BDLB_ROOT_DIR,
                                                    "leaderboard")
  benchmark_dir = os.path.join(leaderboard_dir, benchmark)
  if not os.path.exists(benchmark_dir):
    ValueError("No leaderboard data found at {}".format(benchmark_dir))

  # Metrics for which values are stored
  metrics = os.listdir(benchmark_dir)
  for metric in metrics:
    fig, ax = plt.subplots()
    # Iterate over baselines
    baselines = os.listdir(os.path.join(benchmark_dir, metric))
    for b, baseline in enumerate(baselines):
      baseline = baseline.replace(".csv", "")
      # Fetch results
      df = pd.read_csv(
          os.path.join(benchmark_dir, metric, "{}.csv".format(baseline)))
      # Parse columns
      retained_data = df["retained_data"]
      mean = df["mean"]
      std = df["std"]
      # Visualize mean with standard error
      ax.plot(retained_data,
              mean,
              label=baseline,
              color=COLORS[b % len(COLORS)],
              marker=MARKERS[b % len(MARKERS)])
      ax.fill_between(retained_data,
                      mean - std,
                      mean + std,
                      color=COLORS[b % len(COLORS)],
                      alpha=0.25)
    if results is not None:
      # Plot results from dictionary
      if metric in results:
        df = results[metric]
        baseline = df.name if hasattr(df, "name") else "new_method"
        # Parse columns
        retained_data = df["retained_data"]
        mean = df["mean"]
        std = df["std"]
        # Visualize mean with standard error
        ax.plot(retained_data,
                mean,
                label=baseline,
                color=COLORS[(b + 1) % len(COLORS)],
                marker=MARKERS[(b + 1) % len(MARKERS)])
        ax.fill_between(retained_data,
                        mean - std,
                        mean + std,
                        color=COLORS[(b + 1) % len(COLORS)],
                        alpha=0.25)
    ax.set(xlabel="retained data", ylabel=metric)
    ax.legend()
    fig.tight_layout()
    if isinstance(output_dir, str):
      os.makedirs(output_dir, exist_ok=True)
      fig.savefig(os.path.join(output_dir, "{}.pdf".format(metric)),
                  trasparent=True,
                  dpi=300,
                  format="pdf")
    fig.show()


# Copyright 2019 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Benchmarks registry handlers and definitions."""

from ..diabetic_retinopathy_diagnosis.benchmark import DiabeticRetinopathyDiagnosisBecnhmark

# Internal registry containing <str registered_name, Benchmark subclass>
_BENCHMARK_REGISTRY = {
    "diabetic_retinopathy_diagnosis": DiabeticRetinopathyDiagnosisBecnhmark
}


def load(benchmark,
         level=None,
         data_dir=None,
         download_and_prepare=True,
         **dtask_kwargs):
  """Loads the named benchmark into a `bdlb.Benchmark`.
  
  Args:
    benchmark: `str`, the registerd name of `bdlb.Benchmark`.
    level: `bdlb.Level` or `str` (optional), which level of the
      benchmark to load. If None, will return the realworld level.
    data_dir: `str` (optional), directory to read/write data.
      Defaults to "~/.bdlb/data".
    download_and_prepare: (optional) `bool`, if the data is not available
        it downloads and preprocesses it.
    dtask_kwargs: key arguments for the benchmark contructor.
  
  Returns:
    A registered `bdlb.Benchmark` with `level` at `data_dir`.
  
  Raises:
    BenchmarkNotFoundError: if `name` is unrecognised.
  """
  if not benchmark in _BENCHMARK_REGISTRY:
    raise BenchmarkNotFoundError(benchmark)
  # Fetch benchmark object
  return _BENCHMARK_REGISTRY.get(benchmark)(
      level=level,
      data_dir=data_dir,
      download_and_prepare=download_and_prepare,
      **dtask_kwargs)


class BenchmarkNotFoundError(ValueError):
  """The requested `bdlb.Benchmark` was not found."""

  def __init__(self, name):
    all_denchmarks_str = "\n\t- ".join([""] + list(_BENCHMARK_REGISTRY.keys()))
    error_str = (
        "Benchmark {name} not found. Available denchmarks: {benchmarks}\n",
        format(name=name, benchmarks=all_denchmarks_str))
    super(BenchmarkNotFoundError, self).__init__(error_str)


# Copyright 2018 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Data structures and API for general benchmarks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections


class BenchmarkInfo(
    collections.namedtuple("BenchmarkInfo", [
        "description",
        "urls",
        "setup",
        "citation",
    ])):
  """Abstract class for benchmark information."""
  pass


class DataSplits(
    collections.namedtuple("DataSplits", [
        "train",
        "validation",
        "test",
    ])):
  pass


class Benchmark(object):
  """Abstract class for benchmark objects, specifying the core API."""

  def download_and_prepare(self):
    """Downloads and prepares necessary datasets for benchmark."""
    raise NotImplementedError()

  @property
  def info(self):
    """Text description of the benchmark."""
    raise NotImplementedError()

  @property
  def level(self):
    """The downstream task level."""
    raise NotImplementedError()



# Copyright 2019 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Diabetic retinopathy diagnosis BDL Benchmark."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import collections

from absl import logging

from ..core.levels import Level
from ..core.constants import DATA_DIR
from ..core.benchmark import Benchmark
from ..core.benchmark import BenchmarkInfo
from ..core import transforms

_DIABETIC_RETINOPATHY_DIAGNOSIS_DATA_DIR = os.path.join(
    DATA_DIR, "downloads", "manual", "diabetic_retinopathy_diagnosis")


class DiabeticRetinopathyDiagnosisBecnhmark(Benchmark):
  """Diabetic retinopathy diagnosis benchmark class."""

  def __init__(self,
               level,
               batch_size=64,
               data_dir=None,
               download_and_prepare=False):
    """Constructs a benchmark object.
    
    Args:
      level: `Level` or `str, downstream task level.
      batch_size: (optional) `int`, number of datapoints
        per mini-batch.
      data_dir: (optional) `str`, path to parent data directory.
      download_and_prepare: (optional) `bool`, if the data is not available
        it downloads and preprocesses it.
    """
    self.__level = level if isinstance(level, Level) else Level.from_str(level)
    try:
      self.__ds = self.load(level=level,
                            batch_size=batch_size,
                            data_dir=data_dir or DATA_DIR)
    except AssertionError:
      if not download_and_prepare:
        raise
      else:
        logging.info(
            "Data not found, `DiabeticRetinopathyDiagnosisBecnhmark.download_and_prepare()`"
            " is now running...")
        self.download_and_prepare()

  @classmethod
  def evaluate(cls, estimator, dataset, output_dir=None, name=None):
    """Evaluates an `estimator` on the `mode` benchmark dataset.
    
    Args:
      estimator: `lambda x: mu_x, uncertainty_x`, an uncertainty estimation
        function, which returns `mean_x` and predictive `uncertainty_x`.
      dataset: `tf.data.Dataset`, on which dataset to performance evaluation.
      output_dir: (optional) `str`, directory to save figures.
      name: (optional) `str`, the name of the method.
    """
    import inspect
    import tqdm
    import numpy as np
    import tensorflow_datasets as tfds
    from sklearn.metrics import roc_auc_score
    from sklearn.metrics import accuracy_score
    import matplotlib.pyplot as plt
    COLORS = plt.rcParams['axes.prop_cycle'].by_key()['color']

    # Containers used for caching performance evaluation
    y_true = list()
    y_pred = list()
    y_uncertainty = list()

    # Convert to NumPy iterator if necessary
    ds = dataset if inspect.isgenerator(dataset) else tfds.as_numpy(dataset)

    for x, y in tqdm.tqdm(ds):
      # Sample from probabilistic model
      mean, uncertainty = estimator(x)
      # Cache predictions
      y_true.append(y)
      y_pred.append(mean)
      y_uncertainty.append(uncertainty)

    # Use vectorized NumPy containers
    y_true = np.concatenate(y_true).flatten()
    y_pred = np.concatenate(y_pred).flatten()
    y_uncertainty = np.concatenate(y_uncertainty).flatten()
    fractions = np.asarray([0.5, 0.6, 0.7, 0.8, 0.9, 1.0])

    # Metrics for evaluation
    metrics = zip(["accuracy", "auc"], cls.metrics())

    return {
        metric: cls._evaluate_metric(
            y_true,
            y_pred,
            y_uncertainty,
            fractions,
            lambda y_true, y_pred: metric_fn(y_true, y_pred).numpy(),
            name,
        ) for (metric, metric_fn) in metrics
    }

  @staticmethod
  def _evaluate_metric(y_true,
                       y_pred,
                       y_uncertainty,
                       fractions,
                       metric_fn,
                       name=None):
    """Evaluate model predictive distribution on `metric_fn`
    at data retain `fractions`.
    
    Args:
      y_true: `numpy.ndarray`, the ground truth labels, with shape [N].
      y_pred: `numpy.ndarray`, the model predictions, with shape [N].
      y_uncertainty: `numpy.ndarray`, the model uncertainties,
        with shape [N].
      fractions: `iterable`, the percentages of data to retain for
        calculating `metric_fn`.
      metric_fn: `lambda(y_true, y_pred) -> float`, a metric
        function that provides a score given ground truths
        and predictions.
      name: (optional) `str`, the name of the method.

    Returns:
      A `pandas.DataFrame` with columns ["retained_data", "mean", "std"],
      that summarizes the scores at different data retained fractions.
    """
    import numpy as np
    import pandas as pd

    N = y_true.shape[0]

    # Sorts indexes by ascending uncertainty
    I_uncertainties = np.argsort(y_uncertainty)

    # Score containers
    mean = np.empty_like(fractions)
    # TODO(filangel): do bootstrap sampling and estimate standard error
    std = np.zeros_like(fractions)

    for i, frac in enumerate(fractions):
      # Keep only the %-frac of lowest uncertainties
      I = np.zeros(N, dtype=bool)
      I[I_uncertainties[:int(N * frac)]] = True
      mean[i] = metric_fn(y_true[I], y_pred[I])

    # Store
    df = pd.DataFrame(dict(retained_data=fractions, mean=mean, std=std))
    df.name = name

    return df

  @property
  def datasets(self):
    """Pointer to the processed datasets."""
    return self.__ds

  @property
  def info(self):
    """Text description of the benchmark."""
    return BenchmarkInfo(description="", urls="", setup="", citation="")

  @property
  def level(self):
    """The downstream task level."""
    return self.__level

  @staticmethod
  def loss():
    """Loss used for training binary classifiers."""
    import tensorflow as tf
    tfk = tf.keras
    return tfk.losses.BinaryCrossentropy()

  @staticmethod
  def metrics():
    """Evaluation metrics used for monitoring training."""
    import tensorflow as tf
    tfk = tf.keras
    return [tfk.metrics.BinaryAccuracy(), tfk.metrics.AUC()]

  @staticmethod
  def class_weight():
    """Class weights used for rebalancing the dataset,
    by skewing the `loss` accordingly."""
    return [1.0, 4.0]

  @classmethod
  def load(cls, level="realworld", batch_size=64, data_dir=None,
           as_numpy=False):
    """Loads the datasets for the benchmark.

    Args:
      level: `Level` or `str, downstream task level.
      batch_size: (optional) `int`, number of datapoints
        per mini-batch.
      data_dir: (optional) `str`, path to parent data directory.
      as_numpy: (optional) `bool`, if True returns python generators
        with `numpy.ndarray` outputs.

    Returns:
      A namedtuple with properties:
        * train: `tf.data.Dataset`, train dataset.
        * validation: `tf.data.Dataset`, validation dataset.
        * test: `tf.data.Dataset`, test dataset.
    """
    import numpy as np
    import tensorflow as tf
    import tensorflow_datasets as tfds
    from ..core.benchmark import DataSplits
    from .tfds_adapter import DiabeticRetinopathyDiagnosis

    # Fetch datasets
    try:
      ds_train, ds_validation, ds_test = DiabeticRetinopathyDiagnosis(
          data_dir=data_dir or DATA_DIR,
          config=level).as_dataset(split=["train", "validation", "test"],
                                   shuffle_files=True,
                                   batch_size=batch_size)
    except AssertionError as ae:
      raise AssertionError(
          str(ae) +
          " Run DiabeticRetinopathyDiagnosisBecnhmark.download_and_prepare()"
          " first and then retry.")

    # Parse task level
    level = level if isinstance(level, Level) else Level.from_str(level)
    # Dataset tranformations
    transforms_train, transforms_eval = cls._preprocessors()
    # Apply transformations
    ds_train = ds_train.map(transforms_train,
                            num_parallel_calls=tf.data.experimental.AUTOTUNE)
    ds_validation = ds_validation.map(
        transforms_eval, num_parallel_calls=tf.data.experimental.AUTOTUNE)
    ds_test = ds_test.map(transforms_eval,
                          num_parallel_calls=tf.data.experimental.AUTOTUNE)

    # Prefetches datasets to memory
    ds_train = ds_train.prefetch(tf.data.experimental.AUTOTUNE)
    ds_validation = ds_validation.prefetch(tf.data.experimental.AUTOTUNE)
    ds_test = ds_test.prefetch(tf.data.experimental.AUTOTUNE)

    if as_numpy:
      # Convert to NumPy iterators
      ds_train = tfds.as_numpy(ds_train)
      ds_validation = tfds.as_numpy(ds_validation)
      ds_test = tfds.as_numpy(ds_test)

    return DataSplits(ds_train, ds_validation, ds_test)

  @classmethod
  def download_and_prepare(cls, levels=None):
    """Downloads dataset from Kaggle, extracts zip files
    and processes it using `tensorflow_datasets`.
    
    Args:
      levels: (optional) `iterable` of `str`, specifies which
        levels from {'medium', 'realworld'} to prepare,
        if None it prepares all the levels.

    Raises:
      OSError: if `~/.kaggle/kaggle.json` is not set up. 
    """
    # Disable GPU for data download, extraction and preparation
    import os
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
    cls._download()
    cls._extract()
    cls._prepare(levels)

  @staticmethod
  def _download():
    """Downloads data from Kaggle using `tensorflow_datasets`.
    
    Raises:
      OSError: if `~/.kaggle/kaggle.json` is not set up. 
    """
    import subprocess as sp
    import tensorflow_datasets as tfds

    # Append `/home/$USER/.local/bin` to path
    os.environ["PATH"] += ":/home/{}/.local/bin/".format(os.environ["USER"])

    # Download all files from Kaggle
    drd = tfds.download.kaggle.KaggleCompetitionDownloader(
        "diabetic-retinopathy-detection")
    try:
      for dfile in drd.competition_files:
        drd.download_file(dfile,
                          output_dir=_DIABETIC_RETINOPATHY_DIAGNOSIS_DATA_DIR)
    except sp.CalledProcessError as cpe:
      raise OSError(
          str(cpe) + "." +
          " Make sure you have ~/.kaggle/kaggle.json setup, fetched from the Kaggle website"
          " https://www.kaggle.com/<username>/account -> 'Create New API Key'."
          " Also accept the dataset license by going to"
          " https://www.kaggle.com/c/diabetic-retinopathy-detection/rules"
          " and look for the button 'I Understand and Accept' (make sure when reloading the"
          " page that the button does not pop up again).")

  @staticmethod
  def _extract():
    """Extracts zip files downloaded from Kaggle."""
    import glob
    import tqdm
    import zipfile
    import tempfile

    # Extract train and test original images
    for split in ["train", "test"]:
      # Extract "<split>.zip.00*"" files to "<split>"
      with tempfile.NamedTemporaryFile() as tmp:
        # Concatenate "<split>.zip.00*" to "<split>.zip"
        for fname in tqdm.tqdm(
            sorted(
                glob.glob(
                    os.path.join(_DIABETIC_RETINOPATHY_DIAGNOSIS_DATA_DIR,
                                 "{split}.zip.00*".format(split=split))))):
          # Unzip "<split>.zip" to "<split>"
          with open(fname, "rb") as ztmp:
            tmp.write(ztmp.read())
        with zipfile.ZipFile(tmp) as zfile:
          for image in tqdm.tqdm(iterable=zfile.namelist(),
                                 total=len(zfile.namelist())):
            zfile.extract(member=image,
                          path=_DIABETIC_RETINOPATHY_DIAGNOSIS_DATA_DIR)
      # Delete "<split>.zip.00*" files
      for splitzip in os.listdir(_DIABETIC_RETINOPATHY_DIAGNOSIS_DATA_DIR):
        if "{split}.zip.00".format(split=split) in splitzip:
          os.remove(
              os.path.join(_DIABETIC_RETINOPATHY_DIAGNOSIS_DATA_DIR, splitzip))

    # Extract "sample.zip", "trainLabels.csv.zip"
    for fname in ["sample", "trainLabels.csv"]:
      zfname = os.path.join(_DIABETIC_RETINOPATHY_DIAGNOSIS_DATA_DIR,
                            "{fname}.zip".format(fname=fname))
      with zipfile.ZipFile(zfname) as zfile:
        zfile.extractall(_DIABETIC_RETINOPATHY_DIAGNOSIS_DATA_DIR)
      os.remove(zfname)

  @staticmethod
  def _prepare(levels=None):
    """Generates the TFRecord objects for medium and realworld experiments."""
    import multiprocessing
    from absl import logging
    from .tfds_adapter import DiabeticRetinopathyDiagnosis
    # Hangle each level individually
    for level in levels or ["medium", "realworld"]:
      dtask = DiabeticRetinopathyDiagnosis(data_dir=DATA_DIR, config=level)
      logging.debug("=== Preparing TFRecords for {} ===".format(level))
      dtask.download_and_prepare()

  @classmethod
  def _preprocessors(cls):
    """Applies transformations to the raw data."""
    import numpy as np
    import tensorflow as tf
    import tensorflow_datasets as tfds
    cv2 = tfds.core.lazy_imports.cv2

    # Transformation hyperparameters
    mean = np.asarray([0.42606387, 0.29752496, 0.21309826])
    stddev = np.asarray([0.27662534, 0.20280295, 0.1687619])

    class Parse(transforms.Transform):
      """Parses datapoints from raw `tf.data.Dataset`."""

      def __call__(self, x):
        """Returns `as_supervised` tuple."""
        return x["image"], x["label"]

    class CastX(transforms.Transform):
      """Casts image to `dtype`."""

      def __init__(self, dtype):
        """Constructs a type caster."""
        self.dtype = dtype

      def __call__(self, x, y):
        """Returns casted image (to `dtype`) and its (unchanged) label as tuple."""
        return tf.cast(x, self.dtype), y

    class To01X(transforms.Transform):
      """Rescales image to [min, max]=[0, 1]."""

      def __call__(self, x, y):
        """Returns rescaled image and its (unchanged) label as tuple."""
        return x / 255.0, y

    # Get augmentation schemes
    [augmentation_config,
     no_augmentation_config] = cls._ImageDataGenerator_config()

    # Transformations for train dataset
    transforms_train = transforms.Compose([
        Parse(),
        CastX(tf.float32),
        To01X(),
        transforms.Normalize(mean, stddev),
        # TODO(filangel): hangle batch with ImageDataGenerator
        # transforms.RandomAugment(**augmentation_config),
    ])

    # Transformations for validation/test dataset
    transforms_eval = transforms.Compose([
        Parse(),
        CastX(tf.float32),
        To01X(),
        transforms.Normalize(mean, stddev),
        # TODO(filangel): hangle batch with ImageDataGenerator
        # transforms.RandomAugment(**no_augmentation_config),
    ])

    return transforms_train, transforms_eval

  @staticmethod
  def _ImageDataGenerator_config():
    """Returns the configs for the `tensorflow.keras.preprocessing.image.ImageDataGenerator`,
    used for the random augmentation of the dataset, following the implementation of
    https://github.com/chleibig/disease-detection/blob/f3401b26aa9b832ff77afe93e3faa342f7d088e5/scripts/inspect_data_augmentation.py."""
    augmentation_config = dict(
        featurewise_center=False,
        samplewise_center=False,
        featurewise_std_normalization=False,
        samplewise_std_normalization=False,
        zca_whitening=False,
        rotation_range=180.0,
        width_shift_range=0.05,
        height_shift_range=0.05,
        shear_range=0.,
        zoom_range=0.10,
        channel_shift_range=0.,
        fill_mode="constant",
        cval=0.,
        horizontal_flip=True,
        vertical_flip=True,
        data_format="channels_last",
    )
    no_augmentation_config = dict(
        featurewise_center=False,
        samplewise_center=False,
        featurewise_std_normalization=False,
        samplewise_std_normalization=False,
        zca_whitening=False,
        rotation_range=0.0,
        width_shift_range=0.0,
        height_shift_range=0.0,
        shear_range=0.,
        zoom_range=0.0,
        channel_shift_range=0.,
        fill_mode="nearest",
        cval=0.,
        horizontal_flip=False,
        vertical_flip=False,
        data_format="channels_last",
    )
    return augmentation_config, no_augmentation_config


# Copyright 2019 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
import io
import csv

import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds
cv2 = tfds.core.lazy_imports.cv2


class DiabeticRetinopathyDiagnosisConfig(tfds.core.BuilderConfig):
  """BuilderConfig for DiabeticRetinopathyDiagnosis."""

  def __init__(self,
               target_height,
               target_width,
               crop=False,
               scale=500,
               **kwargs):
    """BuilderConfig for DiabeticRetinopathyDiagnosis.
    
    Args:
      target_height: `int`, number of pixels in height.
      target_width: `int`, number of pixels in width.
      scale: (optional) `int`, the radius of the neighborhood to apply
        Gaussian blur filtering.
      **kwargs: keyword arguments forward to super.
    """
    super(DiabeticRetinopathyDiagnosisConfig, self).__init__(**kwargs)
    self._target_height = target_height
    self._target_width = target_width
    self._scale = scale

  @property
  def target_height(self):
    return self._target_height

  @property
  def target_width(self):
    return self._target_width

  @property
  def scale(self):
    return self._scale


class DiabeticRetinopathyDiagnosis(tfds.image.DiabeticRetinopathyDetection):

  BUILDER_CONFIGS = [
      DiabeticRetinopathyDiagnosisConfig(
          name="medium",
          version="0.0.1",
          description="Images for Medium level.",
          target_height=256,
          target_width=256,
      ),
      DiabeticRetinopathyDiagnosisConfig(
          name="realworld",
          version="0.0.1",
          description="Images for RealWorld level.",
          target_height=512,
          target_width=512,
      ),
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description="A large set of high-resolution retina images taken under "
        "a variety of imaging conditions. "
        "Ehanced contrast and resized to {}x{}.".format(
            self.builder_config.target_height,
            self.builder_config.target_width),
        features=tfds.features.FeaturesDict({
            "name":
            tfds.features.Text(),  # patient ID + eye. eg: "4_left".
            "image":
            tfds.features.Image(shape=(self.builder_config.target_height,
                                       self.builder_config.target_width, 3)),
            # 0: (no DR)
            # 1: (with DR)
            "label":
            tfds.features.ClassLabel(num_classes=2),
        }),
        urls=["https://www.kaggle.com/c/diabetic-retinopathy-detection/data"],
        citation=tfds.image.diabetic_retinopathy_detection._CITATION,
    )

  def _generate_examples(self, images_dir_path, csv_path=None, csv_usage=None):
    """Yields Example instances from given CSV.
    Applies contrast enhancement as in
    https://github.com/btgraham/SparseConvNet/tree/kaggle_Diabetic_Retinopathy_competition.
    Turns the multiclass (i.e. 5 classes) problem to binary classification according to
    https://www.nature.com/articles/s41598-017-17876-z.pdf.

    Args:
      images_dir_path: path to dir in which images are stored.
      csv_path: optional, path to csv file with two columns: name of image and
        label. If not provided, just scan image directory, don't set labels.
      csv_usage: optional, subset of examples from the csv file to use based on
        the "Usage" column from the csv.
    """
    if csv_path:
      with tf.io.gfile.GFile(csv_path) as csv_f:
        reader = csv.DictReader(csv_f)
        data = [(row["image"], int(row["level"]))
                for row in reader
                if csv_usage is None or row["Usage"] == csv_usage]
    else:
      data = [(fname[:-5], -1)
              for fname in tf.io.gfile.listdir(images_dir_path)
              if fname.endswith(".jpeg")]
    for name, label in data:
      yield {
          "name":
          name,
          "image":
          self._preprocess(tf.io.gfile.GFile("%s/%s.jpeg" %
                                             (images_dir_path, name),
                                             mode="rb"),
                           target_height=self.builder_config.target_height,
                           target_width=self.builder_config.target_width),
          "label":
          int(label > 1),
      }

  @classmethod
  def _preprocess(cls,
                  image_fobj,
                  target_height,
                  target_width,
                  crop=False,
                  scale=500):
    """Resize an image to have (roughly) the given number of target pixels.

    Args:
      image_fobj: File object containing the original image.
      target_height: `int`, number of pixels in height.
      target_width: `int`, number of pixels in width.
      crops: (optional) `bool`, if True crops the centre of the original
        image t the target size.
      scale: (optional) `int`, the radius of the neighborhood to apply
        Gaussian blur filtering.

    Returns:
      A file object.
    """
    # Decode image using OpenCV2.
    image = cv2.imdecode(np.fromstring(image_fobj.read(), dtype=np.uint8),
                         flags=3)
    try:
      a = cls._get_radius(image, scale)
      b = np.zeros(a.shape)
      cv2.circle(img=b,
                 center=(a.shape[1] // 2, a.shape[0] // 2),
                 radius=int(scale * 0.9),
                 color=(1, 1, 1),
                 thickness=-1,
                 lineType=8,
                 shift=0)
      image = cv2.addWeighted(src1=a,
                              alpha=4,
                              src2=cv2.GaussianBlur(
                                  src=a, ksize=(0, 0), sigmaX=scale // 30),
                              beta=-4,
                              gamma=128) * b + 128 * (1 - b)
    except cv2.error:
      pass
    # Reshape image to target size
    image = cv2.resize(image, (target_height, target_width))
    # Encode the image with quality=72 and store it in a BytesIO object.
    _, buff = cv2.imencode(".jpg", image, [int(cv2.IMWRITE_JPEG_QUALITY), 72])
    return io.BytesIO(buff.tostring())

  @staticmethod
  def _get_radius(img, scale):
    """Returns radius of the circle to use.
      
    Args:
      img: `numpy.ndarray`, an image, with shape [height, width, 3].
      scale: `int`, the radius of the neighborhood.
    
    Returns:
      A resized image.
    """
    x = img[img.shape[0] // 2, ...].sum(axis=1)
    r = 0.5 * (x > x.mean() // 10).sum()
    s = scale * 1.0 / r
    return cv2.resize(img, (0, 0), fx=s, fy=s)



# Copyright 2019 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Data augmentation and transformations."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


class Transform(object):
  """Abstract transformation class."""

  def __call__(self):
    raise NotImplementedError()


class Compose(Transform):
  """Uber transformation, composing a list of transformations."""

  def __init__(self, trans):
    """Constructs a composition of transformations.
    
    Args:
      trans: `iterable`, sequence of transformations to be composed.
    """
    self.trans = trans

  def __call__(self, x):
    """Returns a composite function of transformations.
    
    Args:
      x: `any`, raw data format.
    
    Returns:
      A composite function to be used with `tf.data.Dataset.map()`.
    """
    import inspect
    for f in self.trans:
      nargs = len(inspect.signature(f).parameters)
      if "y" in locals():
        x, y = f(x, y)
      else:
        if nargs == 1:
          x = f(x)
        else:
          x, y = f(x[0], x[1])
    if "y" in locals():
      return x, y
    else:
      return x


class RandomAugment(Transform):

  def __init__(self, **config):
    """Constructs a tranformer from `config`.
    
    Args:
      **config: keyword arguments for
        `tensorflow.keras.preprocessing.image.ImageDataGenerator`
    """
    self.idg = tf.keras.preprocessing.image.ImageDataGenerator(**config)

  def __call__(self, x, y):
    """Returns a randomly augmented image and its label.
    
    Args:
      x: `tensorflow.Tensor`, an image, with shape [height, width, channels].
      y: `tensorflow.Tensor`, a target, with shape [].
    
    
    Returns:
      The processed tuple:
        * `x`: `tensorflow.Tensor`, the randomly augmented image,
          with shape [height, width, channels].
        * `y`: `tensorflow.Tensor`, the unchanged target, with shape [].
    """
    return tf.py_function(self._transform, inp=[x], Tout=tf.float32), y

  def _transform(self, x):
    """Helper function for `tensorflow.py_function`, will
    be removed when TensorFlow 2.0 is released."""
    return tf.cast(self.idg.random_transform(x.numpy()), tf.float32)


class Resize(Transform):

  def __init__(self, target_height, target_width):
    """Constructs an image resizer.
    
    Args:
      target_height: `int`, number of pixels in height.
      target_width: `int`, number of pixels in width.
    """
    self.target_height = target_height
    self.target_width = target_width

  def __call__(self, x, y):
    """Returns a resized image."""
    return tf.image.resize(x, size=[self.target_height, self.target_width]), y


class Normalize(Transform):

  def __init__(self, loc, scale):
    self.loc = loc
    self.scale = scale

  def __call__(self, x, y):
    return (x - self.loc) / self.scale, y


# Copyright 2019 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
  long_description = f.read()

setup(
    name="bdl-benchmarks",
    version="0.0.1",
    description="BDL Benchmarks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/oatml/bdl-benchmarks",
    author="Oxford Applied and Theoretical Machine Learning Group",
    author_email="oatml@googlegroups.com",
    license="Apache-2.0",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "scipy",
        "pandas",
        "matplotlib",
        "seaborn",
        "scikit-learn",
        "kaggle",
        "tensorflow-gpu==2.0.0-beta0",
        "tfp-nightly",
        "tensorflow-datasets",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



# Copyright 2019 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Downstream tasks levels."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import enum


class Level(enum.IntEnum):
  """Downstream task levels.
  
  TOY: Fewer examples and drastically reduced input dimensionality.
    This version is intended for sanity checks and debugging only.
    Training on a modern CPU should take five to ten minutes.
  MEDIUM: Fewer examples and reduced input dimensionality.
    This version is intended for prototyping before moving on to the
    real-world scale data. Training on a single modern GPU should take
    five or six hours.
  REALWORLD: The full dataset and input dimensionality. This version is
    intended for the evaluation of proposed techniques at a scale applicable
    to the real world. There are no guidelines for train time for the real-world
    version of the task, reflecting the fact that any improvement will translate to
    safer, more robust and reliable systems.
  """

  TOY = 0
  MEDIUM = 1
  REALWORLD = 2

  @classmethod
  def from_str(cls, strvalue):
    """Parses a string value (`strvalue`) to ``Level``."""
    strvalue = strvalue.lower()
    if strvalue == "toy":
      return cls.TOY
    elif strvalue == "medium":
      return cls.MEDIUM
    elif strvalue == "realworld":
      return cls.REALWORLD
    else:
      raise ValueError(
          "Unrecognized level value '{}' provided.".format(strvalue))



# Copyright 2019 BDL Benchmarks Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Default values for some parameters of the API when no values are passed."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

# Root directory of the BDL Benchmarks module.
BDLB_ROOT_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir))

# Directory where to store processed datasets.
DATA_DIR = os.path.join(BDLB_ROOT_DIR, "data")

# URL to hosted datasets
DIABETIC_RETINOPATHY_DIAGNOSIS_URL_MEDIUM = "https://drive.google.com/uc?id=1WAvS-pQsVLxUJiClmKLnVNQkoKmRt2I_"

# URL to hosted assets
LEADERBOARD_DIR_URL = "https://drive.google.com/uc?id=1LQeAfqMQa4lot09qAuzWa3t6attmCeG-"
